@@ -6,7 +6,6 @@ After:
     Transactions on Pattern Analysis and Machine Intelligence 20(5): 522-532.
 """
 
-
 from __future__ import annotations
 
 import abc
@@ -15,9 +14,7 @@ import pickle
 from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple
 
 import numpy
-
 import tqdm
-from scipy import special
 
 from . import actions, util
 
@@ -70,7 +67,7 @@ class ParamDict:
         for vs in (self.delta_sub, self.delta_ins, self.delta_del):
             values.extend(vs.values())
         # Uses sum of exponentiation to maintain logarithmic values.
-        return special.logsumexp(values)
+        return numpy.logaddexp.reduce(values)
 
     @classmethod
     def from_params(cls, other: ParamDict) -> ParamDict:
@@ -88,17 +85,21 @@ class ParamDict:
         self.delta_eos = other.delta_eos
 
     def write_params(self, filepath: str) -> None:
+        util.log_info(f"Writing parameters to {filepath}")
         with open(filepath, "wb") as file:
             pickle.dump(self, file)
 
     @classmethod
     def read_params(cls, filepath: str) -> ParamDict:
-        with open(filepath, "wb") as file:
+        util.log_info(f"Restoring parameters from {filepath}")
+        with open(filepath, "rb") as file:
             return pickle.load(file)
 
 
 class StochasticEditDistance(abc.ABC):
     params: ParamDict
+    _train_progress_bar: Optional[tqdm.tqdm]
+    _val_progress_bar: Optional[tqdm.tqdm]
 
     def __init__(self, params):
         """SED model.
@@ -116,6 +117,8 @@ class StochasticEditDistance(abc.ABC):
         psum = self.params.sum()
         if not numpy.isclose(0.0, psum):
             raise SEDParameterError(f"Parameters do not sum to 1: {psum:.4f}")
+        self._train_progress_bar = None
+        self._val_progress_bar = None
 
     @classmethod
     def build_sed(
@@ -246,7 +249,7 @@ class StochasticEditDistance(abc.ABC):
                         )
                         + alpha[t - 1, v - 1]
                     )
-                alpha[t, v] = special.logsumexp(summands)
+                alpha[t, v] = numpy.logaddexp.reduce(summands)
         alpha[T, V] += self.params.delta_eos
         return alpha
 
@@ -290,7 +293,7 @@ class StochasticEditDistance(abc.ABC):
                         )
                         + beta[t + 1, v + 1]
                     )
-                beta[t, v] = special.logsumexp(summands)
+                beta[t, v] = numpy.logaddexp.reduce(summands)
         return beta
 
     def log_likelihood(
@@ -299,13 +302,14 @@ class StochasticEditDistance(abc.ABC):
         targets: Iterable[Sequence[Any]],
     ) -> float:
         """Computes log likelihood."""
-        ll = numpy.mean(
-            [
-                self.forward_evaluate(source, target)[-1, -1]
-                for source, target in zip(sources, targets)
-            ]
-        )
-        return float(ll)
+        loglikes = []
+        self.val_progress_bar.total = len(sources)
+        self.on_validation_start()
+        for source, target in zip(sources, targets):
+            loglikes.append(self.forward_evaluate(source, target)[-1, -1])
+            self.on_validation_step_end()
+        self.on_validation_epoch_end()
+        return numpy.mean(loglikes)
 
     def em(
         self,
@@ -320,18 +324,19 @@ class StochasticEditDistance(abc.ABC):
             targets (Sequence[Any]): target strings.
             epochs (int): number of EM epochs.
         """
-        loglike = self.log_likelihood(sources, targets)
         gammas = ParamDict.from_params(self.params)
+        self.train_progress_bar.total = len(sources)
         for epoch in range(epochs):
-            with tqdm.tqdm(zip(sources, targets), total=len(sources)) as pbar:
-                pbar.set_description(f"Epoch {epoch}")
-                pbar.set_postfix(loglike=loglike)
-                for source, target in pbar:
-                    self.e_step(source, target, gammas)  # Updates gammas.
-                self.m_step(gammas)  # Updates gammas.
-                self.params.update_params(gammas)  # Updates model parameters.
-                loglike = self.log_likelihood(sources, targets)
-        util.log_info(f"Final log-likelihood: {loglike:.4f}")
+            self.on_train_epoch_start(epoch)
+            for source, target in zip(sources, targets):
+                self.e_step(source, target, gammas)  # Updates gammas.
+                self.on_train_step_end()
+            self.m_step(gammas)  # Updates gammas.
+            self.params.update_params(gammas)  # Updates model parameters.
+            loglike = self.log_likelihood(sources, targets)
+            self.on_train_epoch_end(loss=-loglike)
+        self.on_train_end()
+        self.on_validation_end()
 
     def e_step(
         self, source: Sequence[Any], target: Sequence[Any], gammas: ParamDict
@@ -345,7 +350,7 @@ class StochasticEditDistance(abc.ABC):
         """
         alpha = self.forward_evaluate(source, target)
         beta = self.backward_evaluate(source, target)
-        gammas.delta_eos = special.logsumexp([gammas.delta_eos, 0.0])
+        gammas.delta_eos = numpy.logaddexp(gammas.delta_eos, 0.0)
         T = len(source)
         V = len(target)
         for t in range(T + 1):
@@ -355,7 +360,7 @@ class StochasticEditDistance(abc.ABC):
                 tchar = target[v - 1]
                 stpair = schar, tchar
                 if t > 0 and schar in gammas.delta_del:
-                    gammas.delta_del[schar] = special.logsumexp(
+                    gammas.delta_del[schar] = numpy.logaddexp.reduce(
                         [
                             gammas.delta_del[schar],
                             alpha[t - 1, v]
@@ -364,7 +369,7 @@ class StochasticEditDistance(abc.ABC):
                         ]
                     )
                 if v > 0 and tchar in gammas.delta_ins:
-                    gammas.delta_ins[tchar] = special.logsumexp(
+                    gammas.delta_ins[tchar] = numpy.logaddexp.reduce(
                         [
                             gammas.delta_ins[tchar],
                             alpha[t, v - 1]
@@ -373,7 +378,7 @@ class StochasticEditDistance(abc.ABC):
                         ]
                     )
                 if t > 0 and v > 0 and stpair in gammas.delta_sub:
-                    gammas.delta_sub[stpair] = special.logsumexp(
+                    gammas.delta_sub[stpair] = numpy.logaddexp.reduce(
                         [
                             gammas.delta_sub[stpair],
                             alpha[t - 1, v - 1]
@@ -523,3 +528,56 @@ class StochasticEditDistance(abc.ABC):
                 (action.old, action.old), self.default
             )
         raise SEDActionError(f"Unknown action: {action}")
+
+    # The progress bar implementation is based on:
+    #
+    #     https://github.com/Lightning-AI/pytorch-lightning/
+    #     blob/master/src/lightning/pytorch/callbacks/progress/tqdm_progress.py
+
+    BAR_FORMAT = (
+        "{l_bar}{bar}| {n_fmt}/{total_fmt} "
+        "[{elapsed}<{remaining}, {rate_noinv_fmt}{postfix}]"
+    )
+
+    @property
+    def train_progress_bar(self) -> tqdm.tqdm:
+        if self._train_progress_bar is None:
+            self._train_progress_bar = tqdm.tqdm(
+                position=0, leave=True, bar_format=self.BAR_FORMAT
+            )
+        return self._train_progress_bar
+
+    def on_train_epoch_start(self, epoch: int) -> None:
+        self.train_progress_bar.initial = 0
+        self.train_progress_bar.set_description(f"Epoch {epoch}")
+
+    def on_train_step_end(self) -> None:
+        self.train_progress_bar.update()
+
+    def on_train_epoch_end(self, loss: float) -> None:
+        self.train_progress_bar.set_postfix(loss=loss)
+        self.train_progress_bar.reset()
+
+    def on_train_end(self) -> None:
+        self.train_progress_bar.close()
+
+    @property
+    def val_progress_bar(self) -> tqdm.tqdm:
+        if self._val_progress_bar is None:
+            self._val_progress_bar = tqdm.tqdm(
+                position=1, leave=False, bar_format=self.BAR_FORMAT
+            )
+        return self._val_progress_bar
+
+    def on_validation_start(self) -> None:
+        self.val_progress_bar.initial = 0
+        self.val_progress_bar.set_description("Validating")
+
+    def on_validation_step_end(self) -> None:
+        self.val_progress_bar.update()
+
+    def on_validation_epoch_end(self) -> None:
+        self.val_progress_bar.reset()
+
+    def on_validation_end(self) -> None:
+        self.val_progress_bar.close()
